use crate::entry::Entry;
use crate::poh_recorder::WorkingBankEntry;
use crate::result::Result;
use crate::shred::{Shred, Shredder, RECOMMENDED_FEC_RATE};
use solana_runtime::bank::Bank;
use solana_sdk::signature::Keypair;
use std::sync::mpsc::Receiver;
use std::sync::Arc;
use std::time::{Duration, Instant};

pub(super) struct ReceiveResults {
    pub entries: Vec<Entry>,
    pub time_elapsed: Duration,
    pub bank: Arc<Bank>,
    pub last_tick: u64,
}

/// Theis parameter tunes how many entries are received in one iteration of recv loop
/// This will prevent broadcast stage from consuming more entries, that could have led
/// to delays in shredding, and broadcasting shreds to peer validators
const RECEIVE_ENTRY_COUNT_THRESHOLD: usize = 8;

pub(super) fn recv_slot_entries(receiver: &Receiver<WorkingBankEntry>) -> Result<ReceiveResults> {
    let timer = Duration::new(1, 0);
    let (bank, (entry, mut last_tick)) = receiver.recv_timeout(timer)?;
    let recv_start = Instant::now();

    let mut entries = vec![entry];
    let mut slot = bank.slot();
    let mut max_tick_height = bank.max_tick_height();

    assert!(last_tick <= max_tick_height);

    if last_tick != max_tick_height {
        while let Ok((bank, (entry, tick_height))) = receiver.try_recv() {
            // If the bank changed, that implies the previous slot was interrupted and we do not have to
            // broadcast its entries.
            if bank.slot() != slot {
                entries.clear();
                slot = bank.slot();
                max_tick_height = bank.max_tick_height();
            }
            last_tick = tick_height;
            entries.push(entry);

            if entries.len() >= RECEIVE_ENTRY_COUNT_THRESHOLD {
                break;
            }

            assert!(last_tick <= max_tick_height);
            if last_tick == max_tick_height {
                break;
            }
        }
    }

    let time_elapsed = recv_start.elapsed();
    Ok(ReceiveResults {
        entries,
        time_elapsed,
        bank,
        last_tick,
    })
}

pub(super) fn entries_to_shreds(
    entries: Vec<Entry>,
    last_tick: u64,
    slot: u64,
    bank_max_tick: u64,
    keypair: &Arc<Keypair>,
    latest_shred_index: u64,
    parent_slot: u64,
) -> (Vec<Shred>, u64) {
    let mut shredder = Shredder::new(
        slot,
        parent_slot,
        RECOMMENDED_FEC_RATE,
        keypair,
        latest_shred_index as u32,
    )
    .expect("Expected to create a new shredder");

    bincode::serialize_into(&mut shredder, &entries)
        .expect("Expect to write all entries to shreds");

    if last_tick == bank_max_tick {
        shredder.finalize_slot();
    } else {
        shredder.finalize_data();
    }

    let shred_infos: Vec<Shred> = shredder.shreds.drain(..).collect();

    trace!("Inserting {:?} shreds in blocktree", shred_infos.len());

    (shred_infos, u64::from(shredder.index))
}
